#include "btree_gist.h"
#include "utils/date.h"

typedef struct datekey
{
    DateADT    lower;
    DateADT    upper;
}   DATEKEY;


/*
** date ops
*/
PG_FUNCTION_INFO_V1(gdate_compress);
PG_FUNCTION_INFO_V1(gdate_union);
PG_FUNCTION_INFO_V1(gdate_picksplit);
PG_FUNCTION_INFO_V1(gdate_consistent);
PG_FUNCTION_INFO_V1(gdate_penalty);
PG_FUNCTION_INFO_V1(gdate_same);

Datum    gdate_compress(PG_FUNCTION_ARGS);
Datum    gdate_union(PG_FUNCTION_ARGS);
Datum    gdate_picksplit(PG_FUNCTION_ARGS);
Datum    gdate_consistent(PG_FUNCTION_ARGS);
Datum    gdate_penalty(PG_FUNCTION_ARGS);
Datum    gdate_same(PG_FUNCTION_ARGS);

static void gdate_binary_union(Datum *r1, char *r2);
static int  datekey_cmp(const void *a, const void *b);

/* define for comparison */
#define DATEGE( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    date_ge, \
    DateADTGetDatum( t1 ), \
    DateADTGetDatum( t2 ) \
)))
#define DATEGT( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    date_gt, \
    DateADTGetDatum( t1 ), \
    DateADTGetDatum( t2 ) \
)))
#define DATEEQ( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    date_eq, \
    DateADTGetDatum( t1 ), \
    DateADTGetDatum( t2 ) \
)))
#define DATELT( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    date_lt, \
    DateADTGetDatum( t1 ), \
    DateADTGetDatum( t2 ) \
)))
#define DATELE( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    date_le, \
    DateADTGetDatum( t1 ), \
    DateADTGetDatum( t2 ) \
)))



/**************************************************
 * date ops
 **************************************************/

Datum
gdate_compress(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY  *retval;

  if (entry->leafkey)
  {
    DATEKEY     *r = (DATEKEY *) palloc(sizeof(DATEKEY));

    retval = palloc(sizeof(GISTENTRY));
    r->lower = r->upper = DatumGetDateADT (entry->key);
    gistentryinit(*retval, PointerGetDatum(r),
            entry->rel, entry->page,
            entry->offset, sizeof(DATEKEY), FALSE);
  }
  else
    retval = entry;
  PG_RETURN_POINTER(retval);
}



Datum
gdate_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  DateADT    query  = PG_GETARG_DATEADT( 1 );
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);

  bool        retval ;
  DATEKEY     *key   ;

  if (!entry->key)  return FALSE;
  key = (DATEKEY *) DatumGetPointer(entry->key);

  switch (strategy)
  {
    case BTLessEqualStrategyNumber:
      retval = DATEGE( query, key->lower );
      break;
    case BTLessStrategyNumber:
      if (GIST_LEAF(entry))
        retval = DATEGT(query, key->lower);
      else
        retval = DATEGE(query, key->lower);
      break;
    case BTEqualStrategyNumber:
      /* in leaf page key->lower always = key->upper */
      if (GIST_LEAF(entry))
        retval = DATEEQ(query, key->lower);
      else
        retval = (DATELE( key->lower, query) && DATELE(query, key->upper));
      break;
    case BTGreaterStrategyNumber:
      if (GIST_LEAF(entry))
        retval = DATELT(query, key->upper);
      else
        retval = DATELE(query, key->upper);
      break;
    case BTGreaterEqualStrategyNumber:
      retval = DATELE( query, key->upper );
      break;
    default:
      retval = FALSE;
  }

  PG_RETURN_BOOL( retval );
}



Datum
gdate_union(PG_FUNCTION_ARGS)
{
  bytea     *entryvec = (bytea *) PG_GETARG_POINTER(0);
  int      i,
        numranges;
  DATEKEY     *cur,
       *out = palloc(sizeof(DATEKEY));

  numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
  *(int *) PG_GETARG_POINTER(1) = sizeof(DATEKEY);

  cur = (DATEKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
  out->lower = cur->lower;
  out->upper = cur->upper;

  for (i = 1; i < numranges; i++)
  {
    cur = (DATEKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
    if (DATEGT(out->lower, cur->lower))
      out->lower = cur->lower;
    if (DATELT(out->upper, cur->upper))
      out->upper = cur->upper;
  }

  PG_RETURN_POINTER(out);
}

Datum
gdate_penalty(PG_FUNCTION_ARGS)
{
        DATEKEY      *origentry = (DATEKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
        DATEKEY      *newentry  = (DATEKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
        float           *result = (float *)   PG_GETARG_POINTER(2);
        int32        diff;

        diff = DatumGetInt32(DirectFunctionCall2(
                  date_mi,
                  DateADTGetDatum(newentry->upper),
                  DateADTGetDatum(origentry->upper)));

        *result  = (float) Max(diff, 0);

        diff = DatumGetInt32(DirectFunctionCall2(
                  date_mi,
                  DateADTGetDatum(origentry->lower),
                  DateADTGetDatum(newentry->lower)));

        *result += (float) Max(diff, 0);

        PG_RETURN_POINTER(result);
}

Datum
gdate_picksplit(PG_FUNCTION_ARGS)
{
  PG_RETURN_POINTER(btree_picksplit(
                  (bytea *) PG_GETARG_POINTER(0),
                  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
                  gdate_binary_union,
                  datekey_cmp
                ));
}

Datum
gdate_same(PG_FUNCTION_ARGS)
{
  DATEKEY     *b1 = (DATEKEY *) PG_GETARG_POINTER(0);
  DATEKEY     *b2 = (DATEKEY *) PG_GETARG_POINTER(1);

  bool     *result = (bool *) PG_GETARG_POINTER(2);

  if (b1 && b2)
    *result = (DATEEQ(b1->lower, b2->lower) && DATEEQ(b1->upper, b2->upper)) ? TRUE : FALSE;
  else
    *result = (b1 == NULL && b2 == NULL) ? TRUE : FALSE;
  PG_RETURN_POINTER(result);
}

static void
gdate_binary_union(Datum *r1, char *r2)
{
  DATEKEY     *b1;
  DATEKEY     *b2 = (DATEKEY *) r2;

  if (!DatumGetPointer(*r1))
  {
    *r1 = PointerGetDatum(palloc(sizeof(DATEKEY)));
    b1 = (DATEKEY *) DatumGetPointer(*r1);
    b1->upper = b2->upper;
    b1->lower = b2->lower;
  }
  else
  {
    b1 = (DATEKEY *) DatumGetPointer(*r1);

    b1->lower = (DATEGT(b1->lower, b2->lower)) ?
      b2->lower : b1->lower;
    b1->upper = (DATEGT(b1->upper,b2->upper)) ?
      b1->upper : b2->upper;
  }
}

static int
datekey_cmp(const void *a, const void *b)
{
  return DatumGetInt32(
     DirectFunctionCall2( 
       date_cmp,
             DateADTGetDatum(((DATEKEY *) (((RIX *) a)->r))->lower),
             DateADTGetDatum(((DATEKEY *) (((RIX *) b)->r))->lower)
           )
        );
}

