#include "btree_gist.h"
#include "utils/date.h"

typedef struct timekey
{
    TimeADT    lower;
    TimeADT    upper;
}   TIMEKEY;


/*
** time ops
*/
PG_FUNCTION_INFO_V1(gtime_compress);
PG_FUNCTION_INFO_V1(gtime_tz_compress);
PG_FUNCTION_INFO_V1(gtime_union);
PG_FUNCTION_INFO_V1(gtime_picksplit);
PG_FUNCTION_INFO_V1(gtime_consistent);
PG_FUNCTION_INFO_V1(gtime_tz_consistent);
PG_FUNCTION_INFO_V1(gtime_penalty);
PG_FUNCTION_INFO_V1(gtime_same);

Datum    gtime_compress(PG_FUNCTION_ARGS);
Datum    gtime_tz_compress(PG_FUNCTION_ARGS);
Datum    gtime_union(PG_FUNCTION_ARGS);
Datum    gtime_picksplit(PG_FUNCTION_ARGS);
Datum    gtime_consistent(PG_FUNCTION_ARGS);
Datum    gtime_tz_consistent(PG_FUNCTION_ARGS);
Datum    gtime_penalty(PG_FUNCTION_ARGS);
Datum    gtime_same(PG_FUNCTION_ARGS);

static void gtime_binary_union(Datum *r1, char *r2);
static int  timekey_cmp(const void *a, const void *b);

/* define for comparison */
#define TIMEGE( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    time_ge, \
    TimeADTGetDatum( t1 ), \
    TimeADTGetDatum( t2 ) \
)))
#define TIMEGT( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    time_gt, \
    TimeADTGetDatum( t1 ), \
    TimeADTGetDatum( t2 ) \
)))
#define TIMEEQ( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    time_eq, \
    TimeADTGetDatum( t1 ), \
    TimeADTGetDatum( t2 ) \
)))
#define TIMELT( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    time_lt, \
    TimeADTGetDatum( t1 ), \
    TimeADTGetDatum( t2 ) \
)))
#define TIMELE( t1, t2 ) (DatumGetBool(DirectFunctionCall2( \
    time_le, \
    TimeADTGetDatum( t1 ), \
    TimeADTGetDatum( t2 ) \
)))



/**************************************************
 * time ops
 **************************************************/

Datum
gtime_compress(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY  *retval;

  if (entry->leafkey)
  {
    TIMEKEY     *r = (TIMEKEY *) palloc(sizeof(TIMEKEY));

    retval = palloc(sizeof(GISTENTRY));
    r->lower = r->upper = DatumGetTimeADT (entry->key);
    gistentryinit(*retval, PointerGetDatum(r),
            entry->rel, entry->page,
            entry->offset, sizeof(TIMEKEY), FALSE);
  }
  else
    retval = entry;
  PG_RETURN_POINTER(retval);
}


Datum
gtime_tz_compress(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY  *retval;

  if (entry->leafkey)
  {
    TIMEKEY     *r  = (TIMEKEY   *) palloc(sizeof(TIMEKEY));
    TimeTzADT  *tz = DatumGetTimeTzADTP(entry->key);

    retval = palloc(sizeof(GISTENTRY));

    // We are using the time + zone only to compress
    r->lower = r->upper = ( tz->time + tz->zone ) ;
    gistentryinit(*retval, PointerGetDatum(r),
            entry->rel, entry->page,
            entry->offset, sizeof(TIMEKEY), FALSE);
  }
  else
    retval = entry;
  PG_RETURN_POINTER(retval);
}


static bool gtime_common_consistent ( GISTENTRY * entry, TimeADT * query,  StrategyNumber strategy ){

  bool        retval ;
  TIMEKEY     *key   ;

  if (!entry->key)  return FALSE;
  key = (TIMEKEY *) DatumGetPointer(entry->key);

  switch (strategy)
  {
    case BTLessEqualStrategyNumber:
      retval = TIMEGE( *query, key->lower );
      break;
    case BTLessStrategyNumber:
      if (GIST_LEAF(entry))
        retval = TIMEGT(*query, key->lower);
      else
        retval = TIMEGE(*query, key->lower);
      break;
    case BTEqualStrategyNumber:
      /* in leaf page key->lower always = key->upper */
      if (GIST_LEAF(entry))
        retval = TIMEEQ(*query, key->lower);
      else
        retval = (TIMELE( key->lower, *query) && TIMELE(*query, key->upper));
      break;
    case BTGreaterStrategyNumber:
      if (GIST_LEAF(entry))
        retval = TIMELT(*query, key->upper);
      else
        retval = TIMELE(*query, key->upper);
      break;
    case BTGreaterEqualStrategyNumber:
      retval = TIMELE( *query, key->upper );
      break;
    default:
      retval = FALSE;
  }

  return retval;

}


Datum
gtime_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  TimeADT    query  = PG_GETARG_TIMEADT( 1 );
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
  PG_RETURN_BOOL( gtime_common_consistent ( entry, &query, strategy ) );
}

Datum
gtime_tz_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY  * entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  TimeTzADT  * query = PG_GETARG_TIMETZADT_P( 1 );
  TimeADT      qqq   = query->time + query->zone ;
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
  PG_RETURN_BOOL( gtime_common_consistent ( entry, &qqq, strategy ) );
}



Datum
gtime_union(PG_FUNCTION_ARGS)
{
  bytea     *entryvec = (bytea *) PG_GETARG_POINTER(0);
  int      i,
        numranges;
  TIMEKEY     *cur,
       *out = palloc(sizeof(TIMEKEY));

  numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
  *(int *) PG_GETARG_POINTER(1) = sizeof(TIMEKEY);

  cur = (TIMEKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
  out->lower = cur->lower;
  out->upper = cur->upper;

  for (i = 1; i < numranges; i++)
  {
    cur = (TIMEKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
    if (TIMEGT(out->lower, cur->lower))
      out->lower = cur->lower;
    if (TIMELT(out->upper, cur->upper))
      out->upper = cur->upper;
  }

  PG_RETURN_POINTER(out);
}

Datum
gtime_penalty(PG_FUNCTION_ARGS)
{
        TIMEKEY      *origentry = (TIMEKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
        TIMEKEY      *newentry  = (TIMEKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
        float           *result = (float *)   PG_GETARG_POINTER(2);
        Interval   *intr;

        intr = DatumGetIntervalP(DirectFunctionCall2(
                  time_mi_time,
                  TimeADTGetDatum(newentry->upper),
                  TimeADTGetDatum(origentry->upper)));

        /* see interval_larger */
        *result = Max(intr->time + intr->month * (30.0 * 86400), 0);
        pfree(intr);

        intr = DatumGetIntervalP(DirectFunctionCall2(
                  time_mi_time,
                  TimeADTGetDatum(origentry->lower),
                  TimeADTGetDatum(newentry->lower)));

        /* see interval_larger */
        *result += Max(intr->time + intr->month * (30.0 * 86400), 0);
        pfree(intr);

        PG_RETURN_POINTER(result);
}

Datum
gtime_picksplit(PG_FUNCTION_ARGS)
{
  PG_RETURN_POINTER(btree_picksplit(
                  (bytea *) PG_GETARG_POINTER(0),
                  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
                  gtime_binary_union,
                  timekey_cmp
                ));
}

Datum
gtime_same(PG_FUNCTION_ARGS)
{
  TIMEKEY     *b1 = (TIMEKEY *) PG_GETARG_POINTER(0);
  TIMEKEY     *b2 = (TIMEKEY *) PG_GETARG_POINTER(1);

  bool     *result = (bool *) PG_GETARG_POINTER(2);

  if (b1 && b2)
    *result = (TIMEEQ(b1->lower, b2->lower) && TIMEEQ(b1->upper, b2->upper)) ? TRUE : FALSE;
  else
    *result = (b1 == NULL && b2 == NULL) ? TRUE : FALSE;
  PG_RETURN_POINTER(result);
}

static void
gtime_binary_union(Datum *r1, char *r2)
{
  TIMEKEY     *b1;
  TIMEKEY     *b2 = (TIMEKEY *) r2;

  if (!DatumGetPointer(*r1))
  {
    *r1 = PointerGetDatum(palloc(sizeof(TIMEKEY)));
    b1 = (TIMEKEY *) DatumGetPointer(*r1);
    b1->upper = b2->upper;
    b1->lower = b2->lower;
  }
  else
  {
    b1 = (TIMEKEY *) DatumGetPointer(*r1);

    b1->lower = (TIMEGT(b1->lower, b2->lower)) ?
      b2->lower : b1->lower;
    b1->upper = (TIMEGT(b1->upper,b2->upper)) ?
      b1->upper : b2->upper;
  }
}

static int
timekey_cmp(const void *a, const void *b)
{
  return DatumGetInt32(
     DirectFunctionCall2( 
       time_cmp,
             TimeADTGetDatum(((TIMEKEY *) (((RIX *) a)->r))->lower),
             TimeADTGetDatum(((TIMEKEY *) (((RIX *) b)->r))->lower)
           )
        );
}

